/*
 * Copyright (c) 2003, 2004 The University of Wroclaw.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *    1. Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *    3. The name of the University may not be used to endorse or promote
 *       products derived from this software without specific prior
 *       written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
 * NO EVENT SHALL THE UNIVERSITY BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

using Nemerle.Compiler;
using Nemerle.Compiler.Typedtree;

namespace Nemerle.Compiler {

public type MemberTypes = System.Reflection.MemberTypes;
public type BindingFlags = System.Reflection.BindingFlags;
  
public variant MemberKind
{
  | MK_field { field : IField; }
  | MK_method { method : IMethod; }
  | MK_property { prop : IProperty; }
  | MK_type { tycon : Tycon; }
  | MK_event { body : IEvent; }
}

public variant TyconKind
{
  | TK_NemerleType { exact : TypeBuilder; }
  | TK_ExternalNetType { exact : System.Type; }
  | TK_ExternalNemerleType { exact : System.Type; }
}

public variant Accessibility
{
  | A_public
  | A_internal
  | A_protected_or_internal
  | A_protected_and_internal
  | A_protected
  | A_private
}

public interface IMember
{
  GetMemType () : Type;
  GetKind () : MemberKind;
  GetLocation () : Location;
  GetId () : int;
  GetHandle () : System.Reflection.MemberInfo;
  CanAccess (source : Tycon) : bool;

  /* public properties */
  DeclaringType : Tycon {get;}
  Name : string {get;}
  MemberType : MemberTypes {get;}
  IsStatic : bool { get; } // types are always static
  HasBeenUsed : bool { get; set; } // for the 'unused' warnings
  
  Attributes : NemerleAttributes { get; }
  
  /* public methods */
  GetModifiers () : Modifiers;

  /* Return name that is unique for given member/overload, that is 
     name for fields, name/argument types for methods etc */
  OverloadName : string { get; }
}

public interface IField : IMember
{
  IsMutable : bool { get; }
  IsVolatile : bool { get; }
  IsLiteral : bool { get; }
  GetValue () : object;
  GetFieldInfo () : System.Reflection.FieldInfo;
}

public interface IEvent : IMember
{
  GetEventInfo () : System.Reflection.EventInfo;
  GetAdder () : IMethod;
  GetRemover () : IMethod;
}

public interface IMethod : IMember
{
  GetFreshType () : Type;
  GetHeader () : Fun_header;
  GetFunKind () : Fun_kind;
  GetMethodBase () : System.Reflection.MethodBase;
  GetMethodInfo () : System.Reflection.MethodInfo;
  GetConstructorInfo () : System.Reflection.ConstructorInfo;
  GetExternName () : option <string>; // None () if not extern
  IsVarArgs : bool { get; }
}

public interface IProperty : IMember
{
  IsIndexer : bool { get; }
  IsMutable : bool { get; }
  GetPropertyInfo () : System.Reflection.PropertyInfo;
  GetGetter () : IMethod;
  GetSetter () : IMethod;
}

public interface Tycon : IMember
{
  /** Return full qualified type name with namespaces.
   
      It probably should contain [+] instead of [.] for nested types.
   */
  FullName : string { get; } // FIXME: this could be with + or . ?

  /** Look for specified member. 

      Semantics of returning base class member is the same as for 
      [System.Type.GetMember] method.
   */
  LookupMember (name : string) : list <IMember>;

  /** Return kind of this tycon. */
  GetTyconKind () : TyconKind;

  /** */
  GetTydecl () : Type_decl;

  /** Construct substitution resulting from applying this type constructor
      to specified arguments.
      
      For example assuming our type parameters are are ('a, 'b), then
      [MakeSubst (t1, t2)] produces substitution { 'a := t1; 'b := t2 }.
   */
  MakeSubst (parms : list <Type>) : Subst;

  FreshSubst () : Subst;

  /** If this tycon subtypes [tycon], then return instantiation using
      which we subtype it.
   */
  SuperType (tycon : Tycon) : option <list <Type>>;

  /** */
  SubtypingSubst (tycon : Tycon) : Subst;

  SuperClass () : option <Tycon>;

  /** Return list of all members defined in this very type and its
      supertypes. */
  GetMembers () : list <IMember>;

  /** Searches for the members defined for the current TypeInfo, using the
      specified binding constraints. */
  GetMembers (bindingAttr : BindingFlags) : list <IMember>;

  /** Searches for the fields defined for the current TypeInfo, using the
      specified binding constraints. */
  GetFields (bindingAttr : BindingFlags) : list <IField>;

  GetConstructors (bindingAttr : BindingFlags) : list <IMethod>;
  
  /** */
  GetSuperTypes () : list <Type>;

  GetDirectSuperTypes () : list <Type>;

  /** Compare types for equality. */
  SameAs (the_other : Tycon) : bool;

  GetSystemType () : System.Type;

  /** If this is a constant variant option, this method returns static
      class member that shall be used instead of calling constructor.
      Otherwise [None ()] is returned. */
  GetConstantObject () : option <IField>;

  IsDelegate : bool { get; }

  DefaultIndexerName : option <string> { get; }
  
  GetAccessibility () : Accessibility;

  /**
   * True if the type will be accessible from an external assembly.
   * This depends on this type's and parent types' accessibility.
   */
  IsExternallyAccessible : bool { get; }
  
  GetLibraryReference () : LibraryReference;

  /** If this is constructor of value type (which is allocated on stack not on heap) */
  IsValueType : bool { get;}
  
  HasAttribute (attribute : Tycon) : bool;
}

} // ns
