/*
 * Copyright (c) 2003, 2004 The University of Wroclaw.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *    1. Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *    3. The name of the University may not be used to endorse or promote
 *       products derived from this software without specific prior
 *       written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
 * NO EVENT SHALL THE UNIVERSITY BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

using Nemerle.Collections;
using Nemerle.IO;

using Nemerle.Compiler;
using Nemerle.Compiler.Tyops;
using Nemerle.Compiler.Typedtree;

using PT = Nemerle.Compiler.Parsetree;

namespace Nemerle.Compiler {
module Typattern {
  
  variant Range {
    | R_omega        // _, x, everything
    | R_none         // nothing
    | R_any_except { lits : list <Literal>; }
    | R_literal { l : Literal; }
    | R_variant_option { ti : Tycon; range : Range; }
    | R_enum { fld : IField; }
    | R_tuple { ranges : list <Range>; }
    | R_record { ranges : list <IField * Range>; }
    
    | R_or { ranges : list <Range>; }
  }

  range_of_pattern (p : Pattern) : Range {
    match (p) {
      | P_underscore => R_omega ()
      | P_as (p, _) => range_of_pattern (p)
      | P_tuple (args) => R_tuple (List.Map (args, range_of_pattern))
      | P_record (args) =>
        R_record (List.Map (args, fun (n, p) { (n, range_of_pattern (p)) }))
      | P_cons (ti, pat) =>
        R_variant_option (ti, range_of_pattern (pat))
      | P_enum (f, _) => R_enum (f)
      | P_literal (l) => R_literal (l)
      | P_has_type => R_none ()
    }
  }

  // debug routine
  string_of_range (r : Range) : string {
    match (r) {
      | R_omega => "_"
      | R_any_except ([L_bool (true)]) => "false"
      | R_any_except ([L_bool (false)]) => "true"
      | R_any_except (lst) => "(anything except " + lst.ToString () + ")"
      | R_variant_option (ti, r) =>
        ti.FullName + " (" + string_of_range (r) + ")"
      | R_enum (f) =>
        f.Name
      | R_tuple (ranges) =>
        def concat (r, acc) {
          if (acc == "") string_of_range (r)
          else
            acc + ", " + string_of_range (r)
        };
        "(" + List.FoldLeft (ranges, "", concat) + ")"
        
      | R_record (ranges) =>
        def concat (x, acc) {
          def (f, r) = x;
          def r' = (f : IField).Name + " = " + string_of_range (r);
          if (acc == "") r'
          else
            acc + "; " + r'
        };
        "{ " + List.FoldLeft (ranges, "", concat) + " }"
        
      | R_none => "[[None]]"
      | R_literal => "[[Literal]]"
      
      | R_or (ranges) =>
        def concat (r : Range, acc : string) : string {
          if (acc == "") string_of_range (r)
          else
            acc + " | " + string_of_range (r)
        };
        "[ " + List.FoldLeft (ranges, "", concat) + " ]"
    }
  }

  // for user error messages, we just present first branch of R_or
  user_string_of_range (r : Range) : string {
    match (r) {
      | R_omega => "_"
      | R_any_except ([L_bool (true)]) => "false"
      | R_any_except ([L_bool (false)]) => "true"
      | R_any_except (lst) => "(anything except " + lst.ToString () + ")"
      | R_variant_option (ti, r) =>
        ti.FullName + " (" + user_string_of_range (r) + ")"
      | R_enum (f) =>
        f.Name
      | R_tuple (ranges) =>
        def concat (r, acc) {
          if (acc == "") user_string_of_range (r)
          else
            acc + ", " + user_string_of_range (r)
        };
        "(" + List.FoldLeft (ranges, "", concat) + ")"
        
      | R_record (ranges) =>
        def concat (x, acc) {
          def (f, r) = x;
          def r' = (f : IField).Name + " = " + user_string_of_range (r);
          if (acc == "") r'
          else
            acc + "; " + r'
        };
        "{ " + List.FoldLeft (ranges, "", concat) + " }"
        
      | R_none => Util.ice ()
      | R_literal => Util.ice ()
      | R_or ([]) => Util.ice ()
      
      | R_or (x :: _) => user_string_of_range (x)
    }
  }

  simplify_range (r : Range) : Range {
  //  Message.debug ("simpl: " + string_of_range (r) );
    def aint_none (r : Range) : bool {
      match (r) {
        | R_none => false
        | _ => true
      }
    };
    match (r) {
      | R_or ([]) => R_none ()
      | R_or (lst) =>
        match (List.Filter (lst, aint_none)) {
          | [] => R_none ()
          | [x] => x
          | l => R_or (l)
        }
      | R_any_except (lst) =>
        if (List.Exists (lst, fun (x) { x matches L_bool (true) }) &&
            List.Exists (lst, fun (x) { x matches L_bool (false) }))
          R_none ()
        else
          r
      | R_variant_option (_, R_none) => R_none ()
      | R_tuple (ranges) =>
        if (List.ForAll (ranges, aint_none)) 
          r
        else
          R_none ()
      | R_record (ranges) =>
        def is_named_none (_, x) {
          match (x) {
            | R_none => true
            | _ => false
          }
        };
        if (List.Exists (ranges, is_named_none))
          R_none ()
        else
          r
      | x => x 
    }
  }
    
  subtract (r1 : Range, r2 : Range) : bool * Range {
    mutable did_something = false;

    // subtract new_r from cur_r
    def do_sub (cur_r : Range, new_r : Range) : Range {
//    Message.debug ("subx: " + string_of_range (cur_r) + "     " + string_of_range (new_r) );
      match ((cur_r, new_r)) {
        | (R_none, _) => R_none ()

        // R_none is generated from (x : foo), so ...
        | (x, R_none) =>
          did_something = true;
          x

        | (_, R_omega) => 
          did_something = true; 
          R_none ()
          
        | (R_omega, R_literal (l)) =>
          did_something = true;
          R_any_except ([l])

        | (R_or (ranges), _) =>
          R_or (List.Map (ranges, fun (r : Range) : Range { sub (r, new_r) }))

        | (R_any_except (lits), R_literal (l)) =>
          // FIXME: check if l is in lits
          did_something = true;
          R_any_except (l :: lits)

        | (R_any_except, _) => Util.ice ()

        | (R_variant_option (ti1, r1), R_variant_option (ti2, r2)) =>
          if (ti1.SameAs (ti2))
            R_variant_option (ti1, sub (r1, r2))
          else
            cur_r

        | (R_enum (f1), R_enum (f2)) =>
          //Message.debug ($ "$f1 equals $f2");
          if (f1.Equals (f2)) {
            //Message.debug ("yes");
            did_something = true;
            R_none ()
          } else cur_r

        | (R_omega, R_enum (f)) =>
          def t = f.DeclaringType;
          //Message.debug ($ "_ - enum $f ($t)");
          if (t.HasAttribute (InternalType.FlagsAttribute_tc)) {
            //Message.debug ("did");
            did_something = true;
            R_omega ()
          } else {
            def ranges =
              List.FoldLeft (t.GetMembers (), [], fun (mem : IMember, acc) {
                match (mem) {
                  | fld : IField when fld.IsLiteral && fld.DeclaringType.SameAs (t) =>
                    R_enum (fld) :: acc
                  | _ => acc
                }
              });
            sub (R_or (ranges), new_r)
          }
          
        | (cur, R_variant_option (ti, _)) =>
          def make_range (ti : Tycon) : Range {
            R_variant_option (ti, cur)
          };
          match (Option.UnSome (ti.SuperClass ()).GetTydecl ()) {
            | TD_variant (opts) =>
              sub (R_or (List.Map (opts, make_range)), new_r)
            | _ =>
              Util.ice ("not variant")
          }

        | (R_variant_option, R_record) => Util.ice ("FIXME")
       
        | (R_omega, R_tuple (ranges)) =>
          def args = List.Map (ranges, fun (_ : Range) : Range { R_omega () });
          sub (R_tuple (args), new_r)
          
        | (R_tuple (r1), R_tuple (r2)) =>
          mutable did = true;
          def make_sub (r1 : Range, r2 : Range) : Range {
            def (d, r) = subtract (r1, r2);
            if (d) () else did = false;
            r
          };
          def results = List.Map2 (r1, r2, make_sub);
          if (did) {
            did_something = true;

            mutable output = [];
            // r1 - r2 = r3
            def dosub (pref : list <Range>, 
                       cur : list <Range>, 
                       after_sub : list <Range>) : void {
              match ((cur, after_sub)) {
                | (x :: xs, y :: ys) => 
                  def args = List.RevAppend (pref, (y :: xs));
                  output = simplify_range (R_tuple (args)) :: output;
                  dosub (x :: pref, xs, ys)
                | ([], []) => ()
                | _ => Util.ice ()
              }
            };
            dosub ([], r1, results);
            R_or (output)
          } else cur_r
        
        | (R_omega, R_record) =>
          sub (R_record ([]), new_r)

        | (R_record (r1), R_record (r2)) =>
//          Message.debug ("ransub: " + string_of_range (cur_r) + "   " + string_of_range (new_r));
          mutable all_did = true;

          def loop (r2, acc) {
            def (name : IField, range) = r2;
            
            def loop2 (acc, l) {
              match (l) {
                | (f : IField, r) :: xs =>
                  if (name.GetId () == f.GetId ()) {
                    def (did, res) = subtract (r, range);
                    if (did) () else all_did = false;
                    List.RevAppend (xs, (f, res) :: acc)
                  } else
                    loop2 ((f, r) :: acc, xs)
                | [] =>
                  (name, sub (R_omega (), range)) :: acc
              }
            };

            simplify_range (R_record (loop2 ([], r1))) :: acc
          };

          def r = R_or (List.FoldLeft (r2, [], loop));

          if (all_did) {
            did_something = true;
            r
          } else cur_r

        | (R_variant_option, R_tuple) => assert (false)
        | (R_variant_option, R_literal) => assert (false)
        | (R_tuple, R_literal) => assert (false)
        | (R_record, R_literal) => assert (false)
        | (R_record, R_tuple) => assert (false)
        | (R_tuple, R_record) => assert (false)
        | (R_literal, _) => assert (false)
        | (_, R_any_except) => assert (false)
        | (_, R_or) => assert (false)
        | (R_enum, _) =>
          Message.debug ("enum - " + string_of_range (new_r));
          assert (false)
        | (_, R_enum) => assert (false)
      }
    }

    and sub (r1 : Range, r2 : Range) : Range {
      simplify_range (do_sub (r1, r2))
    };

    def ret = sub (r1, r2);
//    Message.debug ("sub: " + string_of_range (r1) + " - " + string_of_range (r2) + " = " + string_of_range (ret));

    (did_something, ret)
  }

  
  /**
   * Give warnings about ``pattern matching not exhaustive'' and 
   * ``unused match clause''.
   */
  public check_matching (cases : list <Match_case>) : void
  {
    def loop (m : Match_case, acc : Range) : Range
    {
      def loop2 (patex : Pattern * Expr, acc : Range) : Range
      {
        def (pat, ex) = patex;
        def (did_something, ret) = subtract (acc, range_of_pattern (pat));

        unless (did_something)
          Message.warning (pat.loc, "this match clause is unused");

        match (ex) {
          | E_literal (L_bool (true)) => ret
          | _ => acc
        }
      };

      List.FoldLeft (m.patterns, acc, loop2)
    }

    def remaining = List.FoldLeft (cases, R_omega (), loop);
    match (remaining) {
      | R_none => ()
      | r =>
        Message.warning ("this matching is not exhaustive, example unmatched value:");
        Message.warning (user_string_of_range (r))
    }
  }


  public strip_enums (p : Pattern) : Pattern
  {
    | P_enum (_, l) => 
      P_literal (p.loc, p.ty, l)

    | P_as (pat, decl) =>
      P_as (p.loc, p.ty, strip_enums (pat), decl)
      
    | P_tuple (a) =>
      P_tuple (p.loc, p.ty, List.Map (a, strip_enums))
      
    | P_record (a) =>
      P_record (p.loc, p.ty, List.Map (a, fun (f, p) { (f, strip_enums (p)) }))
      
    | P_cons (name, arg) =>
      P_cons (p.loc, p.ty, name, strip_enums (arg))
      
    | P_has_type
    | P_underscore
    | P_literal => p
  }

  
  public ty_patterns (ctx : TypingContext, 
                      matched_value_type : Type, 
                      patterns : list <PT.Pattern * option <PT.Expr>>) 
                    : list <Pattern * Expr> * TypingContext {
                    
    mutable vars = NemerleMap ();
    mutable common_vars = None ();
    
    def do_walk (t : Type, p : PT.Pattern) : Pattern {
      match (p) {
        | PT.P_variable ({ idl = ["true"] }) =>
          walk (t, PT.P_literal (L_bool (true)))

        | PT.P_variable ({ idl = ["false"] }) =>
          walk (t, PT.P_literal (L_bool (false)))

        | PT.P_variable (n) =>
          walk (t, PT.P_as (PT.P_underscore (), PT.SS_string (n)))

        | PT.P_as (pat, PT.SS_string ({ idl = n; color = c })) =>
          def pat' = walk (t, pat);
          def t' =
            match (pat') {
              | P_cons (ti, _) =>
                def t' = ti.GetMemType () /- ti.FreshSubst ();
                if (t' >:> t) t'
                else Util.ice ()
              | _ => t
            };
          def pt_name = PT.Name (n, c, null);
          when (vars.Member (pt_name))
            Message.error ("pattern variable `" + List.Head (n) + 
                           "' already seen in this pattern");
          def decl =
            match (common_vars) {
              | None =>
                ctx.DefineLocal (List.Head (n), c, D_pattern_variable (), t');
              | Some (cv) =>
                match ((cv : NemerleMap <PT.Name, LocalValue>).Find (pt_name)) {
                  | Some (decl) => 
                    unless (t' >:> decl.ty) {
                      Message.error ($ "pattern variable `$(decl.name)' used "
                                       "to have type $(decl.ty) but now it "
                                       "has type $t'");
                      Message.hint ("rename variable if they have no connection")
                    };
                    decl
                  | None =>
                    ctx.DefineLocal (List.Head (n), c, D_pattern_variable (), t');
                }
            };
          vars = vars.Add (pt_name, decl);
          P_as (pat', decl)
          
        | PT.P_as (_, _) =>
          Message.fatal_error ("$ operator used outside of quotation <[ ... ]> context")

        | PT.P_has_type (n, needed_type) =>
          def is_iface (t) {
            | T_app (tc, _) =>
              match (tc.GetTydecl ()) {
                | TD_interface => true
                | _ => false
              }
            | _ => false
          };
          def needed_type = ctx.parent_type.BindType (ctx.tenv, needed_type);
          if (t >:> needed_type)
            walk (t, n)
          else
            if (needed_type >:> t || is_iface (needed_type) || is_iface (t)) {
              def ti =
                match (needed_type) {
                  | T_app (ti, _) => ti
                  | _ => 
                    Message.fatal_error ($ "invalid type in (x : type) pattern: $needed_type")
                };
              def pat = P_has_type (p.loc, t, ti);
              match (walk (needed_type, n)) {
                | P_as (P_underscore, decl) =>
                  P_as (p.loc, t, pat, decl)
                | P_underscore => pat
                | _ =>
                  Message.fatal_error ("only variable patterns are allowed (here : type)")
              }
            } else
              Message.fatal_error ($ "matched value has type $t while the "
                                     "pattern enforces $needed_type")

        | PT.P_underscore => P_underscore ()

        | PT.P_literal (lit) =>
          def t' = Tyexpr.ty_literal (ctx, lit);
          match (Tyutil.unify (t, t', false)) {
            | Some (s) => 
              Tyvar.global_store (s);
              P_literal (lit)
            | None =>
              Message.fatal_error ($ "literal has type $t' while matched "
                                     "value has type $t")
          }

        | PT.P_tuple (pats) =>
          match (Tyutil.top_expand (t)) {
            | T_app (ti, _) =>
              def mems = ti.GetFields (BindingFlags.DeclaredOnly %|
                                       BindingFlags.Public %|
                                       BindingFlags.Instance);
              def names = List.Map (mems, fun (x : IField) { x.Name });
              match (names) {
                | _ :: _ =>
                  if (List.Length (names) == List.Length (pats)) {
                    def mk (n : string, p : PT.Pattern) : string * PT.Pattern {
                      (n, p)
                    };
                    walk (t, PT.P_record (List.Map2 (names, pats, mk)))
                  } else
                    Message.fatal_error 
                    (sprintf ("pattern matches %d values, while type %s has %d fields", 
                              List.Length (pats), ti.FullName, List.Length (names)))
                | [] =>
                    Message.fatal_error ("pattern matches " + 
                                         (List.Length (pats)).ToString () + 
                                         " values, while type " + ti.FullName + 
                                         " has no fields")
              }
              
            | T_prod (ts) =>
              if (List.Length (ts) == List.Length (pats))
                P_tuple (List.Map2 (ts, pats, walk))
              else
                Message.fatal_error ("pattern matches " + 
                                     (List.Length (pats)).ToString () + 
                                     "-tuples, while " + 
                                     (List.Length (ts)).ToString () + 
                                     "-tuple is required")
            | vart =>
              def free_vars (acc, n) {
                if (n <= 0) acc
                else free_vars (Tyvar.free_variable () :: acc, n - 1)
              };
              def ts = free_vars ([], List.Length (pats));
              if (vart >:> T_prod (ts))
                P_tuple (List.Map2 (ts, pats, walk))
              else
                Message.fatal_error ($ "pattern matches "
                                       "$(List.Length (pats))-tuples, "
                                       "not values of type $t")
          }
          
        | PT.P_record (l) =>
          def fn (name, pat) {
            // just a placeholder with correct type
            def bogus_expr = E_type_enforcement (null, t);
            def loc = Location_stack.top ();
            bogus_expr.loc = loc;
            bogus_expr.expr_ty = t;
            match (Tyexpr.lookup_member (ctx, loc, bogus_expr, name)) {
              | [(E_field_ref (_, fieldname)) as fieldref] =>
                (fieldname, walk (Tyexpr.type_of (fieldref), pat))
              | _ =>
                Message.fatal_error ("name `" + name + "' was expected to be field")
            }
          };
          P_record (List.Map (l, fn))
          
        | PT.P_cons (PT.SS_string (n), p) =>
          def p =
            match (p) {
              | PT.P_underscore
              | PT.P_record
              | PT.P_tuple => p
              | _ => PT.P_tuple ([p])
            };

          if (System.Char.IsUpper (List.Last (n.idl) [0])) {
            def env = n.GetEnv (ctx.env);
            match (env.LookupSymbol (n.idl)) {
              // FIXME: if there is type and enum field under the same name,
              // the type wins, but there should be an error message
              | [Sym_type_member (fld : IField)] 
                when fld.IsLiteral && 
                     (fld.DeclaringType.GetTydecl () matches TD_enum) =>
                unless (p matches PT.P_underscore)
                  Message.error ("a pattern was supplied after enum field name");
                def val = Option.UnSome (ConstantFolder.FieldValueAsLiteral (fld));
                P_enum (fld, val)

              | _ =>
                def ti = env.GetType (n.idl, ctx.parent_type);
                match (ti.GetTydecl ()) {
                  | TD_variant_option =>
                    def parent_tycon = Option.UnSome (ti.SuperClass ());
                    def parent_type = parent_tycon.GetMemType () /- parent_tycon.FreshSubst ();
                    def option_type = ti.GetMemType () /- ti.FreshSubst ();
                    if (option_type >:> parent_type) () else Util.ice ();
                    // we use parent_type for subtyping in case matched value
                    // has yet undefined type
                    if (parent_type >:> t)
                      // but option_type inside subpattern
                      P_cons (ti, walk (option_type, p))
                    else if (option_type >:> t) {
                      Message.warning ($ "matching of $t with $option_type is redundant");
                      Message.hint_once ("refer to fields of this variant option without matching");
                      P_cons (ti, walk (option_type, p))
                    } else
                      Message.fatal_error ($ "variant option $option_type "
                                             "doesn't subtype $t")
                  | _ =>
                    Message.error ("`" + Util.qid_of_list (n.idl) + "' isn't variant case");
                    P_underscore ()
                }
            }
          } else
            match (p) {
              | PT.P_underscore => walk (t, PT.P_variable (n))
              | _ =>
                Message.fatal_error ("`" + Util.qid_of_list (n.idl) +
                                     "' cannot be variant option, "
                                     "since it starts with lowercase letter")
            }

        | PT.P_typed (body) => body
            
        | PT.P_cons (_, _) =>
          Message.fatal_error ("$ operator may appear only within <[ ... ]> quotation")

        | PT.P_quoted (parse_element) =>
          // we use Macros functions to translate quoted element
          def lifted = 
            match (parse_element) {
              | PT.SyntaxExpr (e) => Macros.quoted_expr (ctx, e)
              | PT.SyntaxType (e) => Macros.quoted_type (ctx, e)
              | PT.SyntaxPattern (e) => Macros.quoted_pattern (ctx, e)
              | PT.SyntaxCase (e) => Macros.quoted_case (ctx, e)
              | PT.SyntaxCaseGuard (a, b) => Macros.quoted_case_guard (ctx, (a, b))
              | PT.SyntaxFunDecl (e) => Macros.quoted_fundecl (ctx, e)
              | PT.SyntaxFunParm (e) => Macros.quoted_fparam (ctx, e)
              | PT.SyntaxDeclaration (e) => Macros.quoted_member (ctx, e)                
              | PT.SyntaxTType =>
                Message.fatal_error ("matching over typed types not supported")
              | PT.SyntaxTTycon
              | PT.SyntaxTMethod
              | PT.SyntaxTEvent
              | PT.SyntaxTField
              | PT.SyntaxTProperty
              | PT.SyntaxTFunParm => 
                Util.ice ("strange syntax element appeared in quotation")
            };
          walk (t, Macros.expr_to_pattern (lifted))

        | PT.P_spliced =>
          Message.fatal_error ("$ operator may appear only within <[ ... ]> quotation")

        | PT.P_patt_list =>
          Message.fatal_error (".. arguments list may appear only within <[ ... ]> quotation")
      }
    }
    and walk (t : Type, p : PT.Pattern) : Pattern {
      def r = Util.locate (p.loc, do_walk (t, p));
      r.loc = p.loc;
      when (r.ty == null) r.ty = t; 
      r
    };


    def ty_pats (acc, patterns) {
      match (patterns) {
        | (pat, expr) :: rest =>
          vars = NemerleMap ();
          def typed_pattern = walk (matched_value_type, pat);
          
          def filter (n, d, m : NemerleMap <PT.Name, LocalValue>) {
            if (vars.Member (n))
              m.Add (n, d)
            else
              m
          };
          
          common_vars =
            match (common_vars) {
              | None => Some (vars)
              | Some (x) => Some (x.Fold (filter, NemerleMap ()))
            };

          def ctx' = ctx.WithLocals (vars.Fold (fun (_, d, acc) { d :: acc }, []));

          def expr =
            match (expr) {
              | None => 
                def ex = E_literal (L_bool (true));
                ex.loc = Location_stack.top ();
                ex
              | Some (x) => 
                def ex = Tyexpr.ty_expr (ctx', x);
                def bool_ty = InternalType.Boolean;
                Tyexpr.expect_type ("when guard", ex, bool_ty);
                ex
            };

          ty_pats ((typed_pattern, expr) :: acc, rest)
          
        | [] =>
          def addvar (_, d : LocalValue, acc) { 
            d.is_pattern_outcome = true; 
            d :: acc
          };
          (List.Rev (acc), ctx.WithLocals ((Option.UnSome (common_vars)).Fold (addvar, [])))
      }
      
    };

    ty_pats ([], patterns)
  }
  

}
}
