/*
 * Copyright (c) 2003, 2004 The University of Wroclaw.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *    1. Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *    3. The name of the University may not be used to endorse or promote
 *       products derived from this software without specific prior
 *       written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
 * NO EVENT SHALL THE UNIVERSITY BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

using Nemerle.Compiler;
using Nemerle.Compiler.Typedtree;

namespace Nemerle.Compiler.Typedtree {

  public class Decl_base : Located
  {
    public mutable name : string;
    public mutable modifiers : NemerleAttributes;
    public mutable id : int;
    public mutable tyinfo : TypeBuilder;

    public this () { this.id = Util.next_id(); }
  }

  [Record]
  public class Constraint
  {
    public tyvar : string;
    public ty : Type;
  }

  public variant Type_decl
  {
    | TD_class
    | TD_alias { ty : Type; }
    | TD_interface
    | TD_variant { members : list <Tycon>; }
    | TD_variant_option
    | TD_enum
  }

  public variant Type
  {
    | T_app {
        tycon : Tycon;
        args : list <Type>;
      }
    | T_var { tyvar : Tyvar.T; }
    | T_ref { t : Type; }
    | T_out { t : Type; }
    | T_fun { from : Type; to : Type; }
    | T_void
    | T_prod { args : list <Type>; }    
    | T_array { t : Type; rank : int; }

    public override ToString () : string
    {
      Tyutil.string_of_type (this)
    }
  }

  [Record]
  public class Fun_parm
  {
    public loc : Location;
    public name : string;
    public color : int;
    public mutable ty : Type;
    public mutable decl : LocalValue;
    public modifiers : Modifiers;
  }

  public enum FunctionUsage 
  {
    | NotUsed
    | UsedJustOnce
    | Used
    | UsedAsFirstClass
  }
  
  public class Fun_header
  {
    public loc : Location;
    public mutable name : string;
    public id : int;
    public mutable ret_type : Type;
    public mutable parms : list <Fun_parm>;
    public mutable typarms : list <Tyvar.T>;
    public tenv : Tyvar.TEnv;

    public mutable usage : FunctionUsage;
    // used for loop generation
    public mutable body : Expr;

    public mutable closure_vars : list <LocalValue>;
    public mutable decl : LocalValue;
    public mutable all_locals : list <LocalValue>;

    public mutable cg_method : CG_method;

    public this (loc : Location, 
                 name : string, 
                 ret_type : Type, 
                 parms : list <Fun_parm>, 
                 typarms : list <Tyvar.T>, 
                 tenv : Tyvar.TEnv)
    {
      this.loc = loc;
      this.name = name;
      this.ret_type = ret_type;
      this.parms = parms;
      this.typarms = typarms;
      this.tenv = tenv;

      id = Util.next_id ();
      closure_vars = [];
      all_locals = [];
      decl = null;
      cg_method = null;
      usage = FunctionUsage.NotUsed;
    }
  }

  [Record]
  public class Function_decl
  {
    public header : Fun_header;
    public body : Expr;
  }

  public class Pattern_base : Located
  {
    public mutable ty : Type;
    public this () {}
    public this (loc : Location, ty : Type) { base (loc); this.ty <- ty; }
  }
  
  public variant Pattern : Pattern_base
  {
    | P_underscore
    | P_as { pat : Pattern; decl : LocalValue; }
    | P_has_type { tycon : Tycon; }
    | P_tuple { args : list <Pattern>; }
    | P_record { args : list <IField * Pattern>; }
    | P_cons { name : Tycon; arg : Pattern; }
    | P_enum { fld : IField; val : Literal; }
    | P_literal { lit : Literal; }
  }

  [Record]
  public class Match_case
  {
    public mutable patterns : list <Pattern * Expr>;
    public body : Expr;
  }
  
  public class Expr_base : Located
  {
    public mutable expr_ty : Type;
    public this () {}
    public this (loc : Location, ty : Type) { this.loc = loc; expr_ty = ty }
  }

  public variant Expr : Expr_base 
  { 
    | E_local_ref { decl : LocalValue; }
    | E_static_ref { mem : IMember; }
    | E_implicit_value_type_ctor { tc : Tycon; }
    | E_field_ref { obj : Expr; fld : IField; }
    | E_constant_object_ref { mem : IField; }
    | E_property_ref { obj : Expr; prop : IProperty; }
    | E_static_property_ref { prop : IProperty; }
    // XXX: notvirtual here is not used anywhere yet, consider removing it
    | E_method_ref { obj : Expr; meth : IMethod; notvirtual : bool; }
    | E_call { func : Expr; parms : list <Parm>; }
    | E_tail_call { func : Expr; parms : list <Parm>; }
    | E_self_tail_call { parms : list <Parm>; }
    | E_assign { target : Expr; source : Expr; }
    | E_let { name : LocalValue; val : Expr; body : Expr; }
    | E_letfun { funs : list <Function_decl>; body : Expr; }
    | E_match { expr : Expr; cases : list <Match_case>; }
    | E_raise { exn : Expr; }
    | E_try_with { body : Expr; exn : LocalValue; handler : Expr; }
    | E_try_finally { body : Expr; handler : Expr; }
    | E_literal { val : Literal; }
    | E_this
    | E_base { base_ctor : IMethod; }
    | E_type_conversion { expr : Expr; target_type : Type; is_checked : bool; }  // (expr :> ty)
    | E_type_enforcement { expr : Expr; target_type : Type; } // (expr : ty)
    | E_sequence { e1 : Expr; e2 : Expr; }
    | E_tuple { args : list <Expr>; }
    | E_mkarray { args : list <Expr>; dimensions : list <int>; }
    | E_typeof { target_type : Type; }
    | E_array_indexer { over_type : Type; obj : Expr; args : list <Expr>; }
    | E_opcode { name : string; }
  }

  public enum ParmKind {
    | Normal
    | Ref
    | Out
  }

  [Record]
  public class Parm
  {
    public kind : ParmKind;
    public expr : Expr;
    public name : string;  // "" if no name given

    public this (expr : Expr)
    {
      kind <- ParmKind.Normal;
      this.expr <- expr;
      name <- "";
    }
  }

 
} // Nemerle.Compiler.Typedtree
